/*
	Database.java
	
	Created: 10/14/2011
	Modified: 12/7/2011
	
	Provides I/O interaction for "Broker Bob" with MySQL database.

	written by: Jakub Kolodziejski
	debugged by: Jakub Kolodziejski
*/

import java.sql.*;
import java.math.*;

public class Database {

	// Controls whether to print debug statements to console/shell
	private boolean debug = true;

	private Connection connection = null;
	private ResultSet orders = null;
	public Order currentOrder = new Order();
	
	// Database connection constructor
	public Database() {	}
	
	// Database connection constructor/connector
	// INPUT:	String	dbName		database name
	//			String	dbUser		database username
	//			String	dbPassword	database password
	public Database(String dbName, String dbUser, String dbPassword) {
		
		connect(dbName, dbUser, dbPassword);
	}
	
	// Connects to a database
	// INPUT:	String	dbName		database name
	//			String	dbUser		database username
	//			String	dbPassword	database password
	public void connect(String dbName, String dbUser, String dbPassword) {
		
		debug("Connecting to database \"" + dbName + "\"...");
		
		try {
			
			// Load the JDBC driver
			String driverName = "com.mysql.jdbc.Driver"; // MySQL JDBC driver
			Class.forName(driverName);
			
			// Create a connection to the database
			String serverName = "localhost";
			String url = "jdbc:mysql://" + serverName +  "/" + dbName; // a JDBC url
			connection = DriverManager.getConnection(url, dbUser, dbPassword);
		}
		// Thrown when JDBC driver cannot be found
		catch (ClassNotFoundException e) { debug("\nClassNotFoundException: " + e.getMessage() + "\n"); }
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
		
		debug("Connected!");
	}
	
	// Reads the orders from orders table and saves to ResultSet orders for later use
	public void readOrders() {
		
		debug("Reading orders from table...");
		
		try {
			
			// Create a result set containing all current open order data
			Statement stmt = connection.createStatement();
			orders = stmt.executeQuery("SELECT * FROM orders ORDER BY time_placed");
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
		
		debug("Finished reading orders.");
	}
	
	// Returns a list of all stock symbols there are open orders for (delimited by a "+" per Yahoo CSV query format)
	// OUTPUT:	String	stock symbols delimited by +
	public String getSymbols() {
		
		debug("Getting list of stock symbols for orders...");
		
		String symbols = "";
		
		try {
			
			// Iterate through list of orders
			while (orders.next()) {
				
				// If symbol is already in the list, skip it.
				if (symbols.indexOf(orders.getString("symbol")) > 0) {
					continue;
				}
				// Add + between symbols if necessary
				if (symbols != "") {
					symbols += "+";
				}
				// Add symbol to return list
				symbols += orders.getString("symbol");
			}
			
			orders.beforeFirst();
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
		
		debug("Returned: " + symbols);
		
		return symbols;		
	}
	
	// Return next order data
	// OUTPUT:	Order	current order (entire row of data)
	public Order getOrder() {
		
		debug("Getting next order data...");
		
		try {
			
			// Fill and return order data structure with data of next order
			if (orders.next()) {
				
				currentOrder.id = orders.getInt("id");
				currentOrder.account = orders.getInt("account_id");
				currentOrder.timePlaced = orders.getTimestamp("time_placed");
				currentOrder.symbol = orders.getString("symbol");
				currentOrder.action = orders.getString("action");
				currentOrder.quantity = orders.getLong("quantity");
				currentOrder.type = orders.getString("type");
				currentOrder.stopPrice = orders.getBigDecimal("stop_price");
				currentOrder.limitPrice = orders.getBigDecimal("limit_price");
				currentOrder.trail = orders.getBigDecimal("trail");
				currentOrder.duration = orders.getString("duration");
				currentOrder.notify = orders.getByte("notify");
				
				debug("Returned: Order(" + currentOrder.id + "," + currentOrder.account + ","
										+ currentOrder.timePlaced + "," + currentOrder.symbol + ","
										+ currentOrder.action + "," + currentOrder.quantity + ","
										+ currentOrder.type + "," + currentOrder.stopPrice + ","
										+ currentOrder.limitPrice + "," + currentOrder.trail + ","
										+ currentOrder.duration + "," + currentOrder.notify + ")");
				
				return currentOrder;
			}
			else {
				debug("No more orders.");
				return null;
			}
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
		
		debug("Returned: NULL");
		
		return null;
	}
	
	// Returns the cash amount of a user
	// INPUT:	int		account		account id
	// OUTPUT:	BigDecimal	cash amount
	public BigDecimal getUserCash(int account){
		
		debug("Getting " + account + "'s cash amount...");
		
		try {
			
			// Create a result set containing cash amount for 'user'
			PreparedStatement prepStmt = connection.prepareStatement("SELECT cash FROM accounts WHERE id=?");
			prepStmt.setInt(1, account);
			ResultSet rs = prepStmt.executeQuery();
			
			// Return cash amount from result set
			if (rs.next()) {
				
				BigDecimal cashMoney = rs.getBigDecimal(1);
				debug("Returned: $" + cashMoney);
				return cashMoney;
			}
			else {
				debug("NO RECORD FOUND!");
			}
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
		
		debug("Returned: ZERO");
		
		return new BigDecimal(0);
	}
	
	// Sets user's cash amount
	// INPUT:	int			account		account id
	//			BigDecimal	cashMoney	new cash amount
	public void setUserCash(int account, BigDecimal cashMoney) {
		
		debug("Setting " + account + "'s cash to $" + cashMoney + " ...");
		
		try {
			
			PreparedStatement prepStmt = connection.prepareStatement("UPDATE accounts SET cash=? WHERE id=?");
			prepStmt.setBigDecimal(1, cashMoney);
			prepStmt.setInt(2, account);
			prepStmt.executeUpdate();
			
			debug("Finished setting cash.");
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
		//finally { prepStmt.close();}
	}
	
	// Return the quantity of a stock the user has in their portfolio
	// INPUT:	int		account 	account id
	//			String	symbol		stock symbol
	// OUTPUT:	long	number of shares of a stock
	public long getPortfolioStock(int account, String symbol) {
		
		debug("Checking " + account + "'s portfolio for " + symbol + " ...");
		
		try {
			
			// Create a result set containing quantity of a stock from user's portfolio
			PreparedStatement prepStmt = connection.prepareStatement("SELECT quantity FROM portfolio WHERE account_id=? AND symbol=?");
			prepStmt.setInt(1, account);
			prepStmt.setString(2, symbol);
			ResultSet rs = prepStmt.executeQuery();
			
			// Return quantity of a stock from result set
			if (rs.next()) {
				
				long quantity = rs.getLong("quantity");
				debug("Returned: " + quantity);
				return quantity;
			}
			else {
				
				debug("NO RECORD FOUND!");
			}
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
		 
		
		// none of the stock owned
		debug("Returned: ZERO");
		return 0;
	}
	
	// Return user's email address
	// INPUT:	int		account		account id
	// OUTPUT:	String	user's email address
	public String getUserEmail(int account) {
		
		debug("Getting " + account + "'s email address...");
		
		try {
			
			// Create a result set containing email address for 'user'
			PreparedStatement prepStmt = connection.prepareStatement("SELECT email FROM accounts WHERE id=?");
			prepStmt.setInt(1, account);
			ResultSet rs = prepStmt.executeQuery();
			
			// Return email address from result set
			if (rs.next()) {
				
				String email = rs.getString(1);
				debug("Returned: " + email);
				return email;
			}
			else {
				
				debug("NO RECORD FOUND!");
			}
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
		
		debug("Returned: NULL");
		return "";
	}
	
	// Return user's phone number
	// INPUT:	int		account		account id
	// OUTPUT:	String	user's phone number
	public String getUserPhone(int account) {
		
		debug("Getting " + account + "'s phone number...");
		
		try {
			
			// Create a result set containing email address for 'user'
			PreparedStatement prepStmt = connection.prepareStatement("SELECT phone FROM accounts WHERE id=?");
			prepStmt.setInt(1, account);
			ResultSet rs = prepStmt.executeQuery();
			
			// Return email address from result set
			if (rs.next()) {
				
				String phone = rs.getString(1);
				debug("Returned: " + phone);
				return phone;
			}
			else {
				
				debug("NO RECORD FOUND!");
			}
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
		
		debug("Returned: NULL");
		return "";
	}
	
	// Return user's phone carrier
	// INPUT:	int		account		account id
	// OUTPUT:	String	user's phone carrier
	public String getUserPhoneCarrier(int account) {
		
		debug("Getting " + account + "'s phone carrier...");
		
		try {
			
			// Create a result set containing email address for 'user'
			PreparedStatement prepStmt = connection.prepareStatement("SELECT carrier FROM accounts WHERE id=?");
			prepStmt.setInt(1, account);
			ResultSet rs = prepStmt.executeQuery();
			
			// Return email address from result set
			if (rs.next()) {
				
				String carrier = rs.getString(1);
				debug("Returned: " + carrier);
				return carrier;
			}
			else {
				
				debug("NO RECORD FOUND!");
			}
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
		
		debug("Returned: NULL");
		return "";
	}
	
	// Return user's SMS address
	// INPUT:	int		account		account id
	// OUTPUT:	String	user's SMS address
	public String getUserSMS(int account) {
		
		debug("Getting " + account + "'s SMS address...");
		
		try {
			
			// Create a result set containing email address for 'user'
			PreparedStatement prepStmt = connection.prepareStatement("SELECT gateway FROM carriers WHERE carrier=?");
			prepStmt.setString(1, getUserPhoneCarrier(account));
			ResultSet rs = prepStmt.executeQuery();
			
			// Return email address from result set
			if (rs.next()) {
				
				String SMSAddress = getUserPhone(account) + rs.getString(1);
				debug("Returned: " + SMSAddress);
				return SMSAddress;
			}
			else {
				
				debug("NO RECORD FOUND!");
			}
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
		
		debug("Returned: NULL");
		return "";
	}
	
	// Write to history table
	// INPUT:	int			account			account id
	//			Timestamp	timePlaced		time the order was placed
	//			String		symbol			stock symbol
	//			String		action			buy/sell
	//			long		quantity		number of shares
	//			String		type 			type of order: market/limit/stop
	//			BigDecimal	stopPrice		stop price if stop/stoplimit order
	//			BigDecimal	limitPrice		limit price if limit/stoplimit order
	//			BigDecimal	trail			trailing amount for trailingstop order
	//			String		duration		time order is good for: day/gtc "good till canelled
	//			BigDecimal	price			price per share order was executed at
	//			BigDecimal	commission		commission charged for order
	//			BigDecimal	orderValue		total value of order (price * number of shares)
	//			BigDecimal	accountValue	total value of account
	public void addHistory(int account, Timestamp timePlaced, String symbol, String action, long quantity, String type, BigDecimal stopPrice, BigDecimal limitPrice, BigDecimal trail, String duration, BigDecimal price, BigDecimal commission, BigDecimal orderValue, BigDecimal accountValue) {
		
		debug("Writing to history table: " + account + "," + timePlaced + "," + symbol + "," + action + ","
											+ quantity + "," + type + "," + stopPrice + "," + limitPrice + ","
											+ trail + ","+ duration + "," + price + "," + commission + ","
											+ orderValue + "," + accountValue + " ...");
		
		try {
			java.util.Date today = new java.util.Date();
			Timestamp timeExecuted = new Timestamp(today.getTime());
			
			PreparedStatement prepStmt = connection.prepareStatement("INSERT INTO history (account_id,time_placed,symbol,action,quantity,type,stop_price,limit_price,trail,duration,price,commission,order_value,account_value,time_executed) VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)");
			prepStmt.setInt(1, account);
			prepStmt.setTimestamp(2, timePlaced);
			prepStmt.setString(3, symbol);
			prepStmt.setString(4, action);
			prepStmt.setLong(5, quantity);
			prepStmt.setString(6, type);
			prepStmt.setBigDecimal(7, stopPrice);
			prepStmt.setBigDecimal(8, limitPrice);
			prepStmt.setBigDecimal(9, trail);
			prepStmt.setString(10, duration);
			prepStmt.setBigDecimal(11, price);
			prepStmt.setBigDecimal(12, commission);
			prepStmt.setBigDecimal(13, orderValue);
			prepStmt.setBigDecimal(14, accountValue);
			prepStmt.setTimestamp(15, timeExecuted);
			prepStmt.executeUpdate();
			
			debug("Finished writing to history table.");
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
		//finally { prepStmt.close();}
	}
	
	// Add stock to user's portfolio
	// INPUT:	int			account		account id
	//			String		symbol		stock symbol
	//			long		quantity	number of shares
	//			BigDecimal	price		price per share
	public void addToPortfolio(int account, String symbol, long quantity, BigDecimal price) {
		
		debug("Adding to " + account + "'s portfolio: " + symbol + " " + quantity + " $" + price + " ...");
		
		try {
			
			debug("Checking if " + symbol + " is in portfolio");
			
			// Check if stock exists in user's portfolio
			PreparedStatement prepStmt = connection.prepareStatement("SELECT * FROM portfolio WHERE account_id=? AND symbol=?");
			prepStmt.setInt(1, account);
			prepStmt.setString(2, symbol);
			ResultSet ifExist = prepStmt.executeQuery();
			
			// If stock exists in user's portfolio, add to the quantity and update price as weighted average
			if (ifExist.next()) {
				
				debug("Stock exists, adding to current...");
				
				// Get "old" quantity and price of stock in portfolio
				prepStmt = connection.prepareStatement("SELECT quantity,price FROM portfolio WHERE account_id=? AND symbol=?");
				prepStmt.setInt(1, account);
				prepStmt.setString(2, symbol);
				ResultSet old = prepStmt.executeQuery();
				if (old.next()) {
					// Calculate new quantity as old + current order
					long newQuantity = old.getLong("quantity") + quantity;
					// Calculate new price as weighted average of old and current order
					BigDecimal newPrice = weightedAverage(price, quantity, old.getBigDecimal("price"), old.getLong("quantity"));
					
					// Update quantity and price of user's stock in portfolio
					prepStmt = connection.prepareStatement("UPDATE portfolio SET quantity=?,price=? WHERE account_id=? AND symbol=?");
					prepStmt.setLong(1, newQuantity);
					prepStmt.setBigDecimal(2, newPrice);
					prepStmt.setInt(3, account);
					prepStmt.setString(4, symbol);
					prepStmt.executeUpdate();
					debug("Finished adding. New values: " + symbol + " " + newQuantity + " $" + newPrice);
				}
				else {
					
					debug("NO RECORD FOUND!");
				}
				//finally { old.close(); prepStmt.close();}
			}
			
			// Else, append to the portfolio
			else {
				
				debug("Stock does not exist, appending...");
				prepStmt = connection.prepareStatement("INSERT INTO portfolio (account_id,symbol,quantity,price) VALUES(?,?,?,?)");
				prepStmt.setInt(1, account);
				prepStmt.setString(2, symbol);
				prepStmt.setLong(3, quantity);
				prepStmt.setBigDecimal(4, price);
				prepStmt.executeUpdate();
				debug("Finished appending.");
			}
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
	}
	
	// Remove shares from user's portfolio
	// INPUT:	int			account		account id
	//			String		symbol		stock symbol
	//			long		quantity	number of shares
	public void removeFromPortfolio(int account, String symbol, long quantity) {
		
		debug("Removing from " + account + "'s portfolio: " + symbol + " " + quantity + " ...");
		
		try {
			
			// Get current amount of owned stock
			debug("Getting current amount of owned stock...");
			PreparedStatement prepStmt = connection.prepareStatement("SELECT quantity FROM portfolio WHERE account_id=? AND symbol=?");
			prepStmt.setInt(1, account);
			prepStmt.setString(2, symbol);
			ResultSet rs = prepStmt.executeQuery();
			rs.next();
			long oldQuantity = rs.getLong("quantity");
			
			// If less stock is sold than owned, save the difference
			if (oldQuantity > quantity) {
				
				debug("More shares owned (" + oldQuantity + ") than sold (" + quantity + "). Updating quantity...");
				prepStmt = connection.prepareStatement("UPDATE portfolio SET quantity=? WHERE account_id=? AND symbol=?");
				long tmp = oldQuantity - quantity;
				prepStmt.setLong(1, tmp);
				prepStmt.setInt(2, account);
				prepStmt.setString(3, symbol);
				prepStmt.executeUpdate();
				debug("Finished updating stock quantity. New value: " + (oldQuantity - quantity));
			}
			// Else, remove completely from portfolio
			else {
				
				debug("All shares sold. Removing stock...");
				prepStmt = connection.prepareStatement("DELETE FROM portfolio WHERE account_id=? AND symbol=?");
				prepStmt.setInt(1, account);
				prepStmt.setString(2, symbol);
				prepStmt.executeUpdate();
				debug("Finished removing stock.");
			}
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
	}
	
	// Delete an order from orders table based on id number
	// INPUT:	int	id	unique id number of the order to delete
	public void deleteOrder(int id) {
		
		debug("Deleting order: " + id + " ...");
		
		try {
			
			PreparedStatement prepStmt = connection.prepareStatement("DELETE FROM orders WHERE id=?");
			prepStmt.setInt(1, id);
			int deleteCount = prepStmt.executeUpdate();
			
			debug("Finished deleting order.");
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
	}
	
	// Updates the stop price of an trailing stop order
	// INPUT:	int 		id			unique id number of the order to update
	//			BigDecimal	stopPrice	new stop price which = market price + trail amount
	public void updateStopPrice(int id, BigDecimal stopPrice) {
		
		debug("Updating order stop price: " + id + " ...");
		
		try {
			
			PreparedStatement prepStmt = connection.prepareStatement("UPDATE orders SET stop_price=? WHERE id=?");
			prepStmt.setBigDecimal(1, stopPrice);
			prepStmt.setInt(2, id);
			prepStmt.executeUpdate();
			
			debug("Finished updating order.");
		}
		catch (SQLException e) { debug("\nSQLException: " + e.getMessage() + " [ERROR CODE:" + e.getErrorCode() + "]\n"); }
	}
	
	// Returns the weighted average of two prices weighted by their respective stock quantities
	// INPUT:	BigDecimal	price1		1st stock price
	//			long		quantity1	1st stock quantity
	//			BigDecimal	price2		2nd stock price
	//			long		quantity2	2nd stock quantity
	// OUTPUT:	BigDecimal	weighted average
	private BigDecimal weightedAverage(BigDecimal price1, long quantity1long, BigDecimal price2, long quantity2long) {
		
		// Convert quantities to BigDecimals
		BigDecimal quantity1 = new BigDecimal(quantity1long);
		BigDecimal quantity2 = new BigDecimal(quantity2long);
		
		/*
			(price1 * quantity1) + (price2 * quantity2)
			------------------------------------------- = weighted average
			          quantitiy1 + quantity2
		*/
		
		BigDecimal top1 = price1.multiply(quantity1);
		BigDecimal top2 = price2.multiply(quantity2);
		
		BigDecimal top = top1.add(top2);
		BigDecimal bottom = quantity1.add(quantity2);
		
		return top.divide(bottom, 2, RoundingMode.HALF_UP);
	}
	
	// Prints debug statements to console/shell.
	private void debug(String text) {
		if (debug) {
			System.out.println("" + text);
		}
	}
}
